//////////////////////////////////////////////////////////////////////
// app.cpp
//
// "Uninteresting" portion of the rt_Lighting sample consisting of
// application boiler plate code and GUI handling.
//
// Rich Thomson
// Copyright (C) 2002, Pahvant Technologies, all rights reserved.
//
#include <iomanip>
#include <sstream>
#include <vector>

#define STRICT
#include <windows.h>
#include <windowsx.h>

#include <atlbase.h>

#include <D3DX8.h>

#include "D3DApp.h"
#include "D3DFont.h"
#include "DXUtil.h"
#include "resource.h"
#include "rt_Lighting.h"
#include "colorsel.h"

#include "rt/app.h"
#include "rt/hr.h"
#include "rt/mesh.h"
#include "rt/misc.h"
#include "rt/tstring.h"

//-----------------------------------------------------------------------------
// Global access to the app (needed for the global WndProc())
//-----------------------------------------------------------------------------
CMyD3DApplication* g_pApp  = NULL;
HINSTANCE          g_hInst = NULL;


//////////////////////////////////////////////////////////////////////
// enable_menu
//
// Enable or disable a menu item based on a boolean.
//
inline void
enable_menu(HMENU menu, UINT item, bool enabled)
{
    ::EnableMenuItem(menu, item, enabled ? MF_ENABLED : MF_GRAYED);
}



//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
    CMyD3DApplication d3dApp;

    g_pApp  = &d3dApp;
    g_hInst = hInst;

    if( FAILED( d3dApp.Create( hInst ) ) )
        return 0;

    return d3dApp.Run();
}




//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
    : m_teapot_nocolors(),
    m_teapot(),
    m_ambient(D3DCOLOR_XRGB(50, 50, 50)),
    m_background(D3DCOLOR_XRGB(100, 100, 100)),
    m_ambient_source(D3DMCS_MATERIAL),
    m_diffuse_source(D3DMCS_MATERIAL),
    m_specular_source(D3DMCS_MATERIAL),
    m_emissive_source(D3DMCS_MATERIAL),
    m_clone_fvf(0),
    m_shade_mode(D3DSHADE_GOURAUD),
    m_fill_mode(D3DFILL_SOLID),
    m_projection(PT_PERSPECTIVE),
    m_animate_view(false),
    m_light_enabled(true),
    m_show_text(true),
    m_lighting(true),
    m_local_viewer(true),
    m_normalize_normals(true),
    m_specular_enable(true),
    m_dithering(true),
    m_color_vertex(true)
{
    m_dwCreationWidth           = 500;
    m_dwCreationHeight          = 375;
    m_strWindowTitle            = TEXT( "rt_Lighting" );
    m_bUseDepthBuffer           = TRUE;

    // Create a D3D font using d3dfont.cpp
    m_pFont                     = new CD3DFont( _T("Arial"), 12, D3DFONT_BOLD );
    m_bLoadingApp               = TRUE;

    ZeroMemory( &m_UserInput, sizeof(m_UserInput) );
    m_fWorldRotX                = 0.0f;
    m_fWorldRotY                = 0.0f;

    const D3DMATERIAL8 mat =
    {
        1, 1, 1, 1,     // diffuse
        0.1f, 0.1f, 0.1f, 0.1f,     // ambient
        1, 1, 1, 1,     // specular
        0, 0, 0, 0,     // emissive
        100             // specular power
    };
    m_material = mat;
    const D3DLIGHT8 light =
    {
        D3DLIGHT_DIRECTIONAL,   // type
        1, 1, 1, 1,             // diffuse
        1, 1, 1, 1,             // specular
        1, 1, 1, 1,             // ambient
        0, 0, -5,               // position
        0, 0, 1,                // direction
        10,                     // cutoff range
        1,                      // falloff
        1, 0, 0,                // attenuation
        0, D3DX_PI/2            // theta, phi
    };
    m_light = light;

    // Read settings from registry
    ReadSettings();
}




//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
    // Drawing loading status message until app finishes loading
    SendMessage( m_hWnd, WM_PAINT, 0, 0 );

    m_bLoadingApp = FALSE;

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: ReadSettings()
// Desc: Read the app settings from the registry
//-----------------------------------------------------------------------------
VOID CMyD3DApplication::ReadSettings()
{
    HKEY hkey;
    if( ERROR_SUCCESS == RegCreateKeyEx( HKEY_CURRENT_USER, DXAPP_KEY, 
        0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hkey, NULL ) )
    {
        // Read the stored window width/height.  This is just an example,
        // of how to use DXUtil_Read*() functions.
        DXUtil_ReadIntRegKey( hkey, TEXT("Width"), &m_dwCreationWidth, m_dwCreationWidth );
        DXUtil_ReadIntRegKey( hkey, TEXT("Height"), &m_dwCreationHeight, m_dwCreationHeight );

        RegCloseKey( hkey );
    }
}




//-----------------------------------------------------------------------------
// Name: WriteSettings()
// Desc: Write the app settings to the registry
//-----------------------------------------------------------------------------
VOID CMyD3DApplication::WriteSettings()
{
    HKEY hkey;
    DWORD dwType = REG_DWORD;
    DWORD dwLength = sizeof(DWORD);

    if( ERROR_SUCCESS == RegCreateKeyEx( HKEY_CURRENT_USER, DXAPP_KEY, 
        0, NULL, REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hkey, NULL ) )
    {
        // Write the window width/height.  This is just an example,
        // of how to use DXUtil_Write*() functions.
        DXUtil_WriteIntRegKey( hkey, TEXT("Width"), m_rcWindowClient.right );
        DXUtil_WriteIntRegKey( hkey, TEXT("Height"), m_rcWindowClient.bottom );

        RegCloseKey( hkey );
    }
}





//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the display device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior,
                                          D3DFORMAT Format )
{
    BOOL bCapsAcceptable;

    bCapsAcceptable = TRUE;

    if( bCapsAcceptable )         
        return S_OK;
    else
        return E_FAIL;
}




//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{

    // Init the font
    m_pFont->InitDeviceObjects( m_pd3dDevice );

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restores scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
    // Set up our view matrix. A view matrix can be defined given an eye point,
    // a point to lookat, and a direction for which way is up. Here, we set the
    // eye five units back along the z-axis and up three units, look at the
    // origin, and define "up" to be in the y-direction.
    D3DXMATRIX matView;
    D3DXVECTOR3 vFromPt   = D3DXVECTOR3( 0.0f, 0.0f, -5.0f );
    D3DXVECTOR3 vLookatPt = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
    D3DXVECTOR3 vUpVec    = D3DXVECTOR3( 0.0f, 1.0f, 0.0f );
    D3DXMatrixLookAtLH( &matView, &vFromPt, &vLookatPt, &vUpVec );
    m_pd3dDevice->SetTransform( D3DTS_VIEW, &matView );

    // Set the projection matrix
    set_projection();

    // Restore the font
    m_pFont->RestoreDeviceObjects();

    {
        // create a teapot mesh
        rt::dx_buffer<DWORD> adj_tmp;
        CComPtr<ID3DXMesh> tmp;
        THR(::D3DXCreateTeapot(m_pd3dDevice, &tmp, &adj_tmp));

        // tesselate it using the N-patch algorithm to get smoother
        // lighting than with the plain teapot.
        rt::dx_buffer<DWORD> adj;
        THR(::D3DXTessellateNPatches(tmp, adj_tmp, 3.0f, true,
            &m_teapot_nocolors, &adj));

        // now optimize it for rendering
        std::vector<DWORD> adj_out(m_teapot_nocolors->GetNumFaces()*3);
        std::vector<DWORD> face_remap(m_teapot_nocolors->GetNumFaces());
        CComPtr<ID3DXBuffer> vertex_remap;
        THR(m_teapot_nocolors->OptimizeInplace(D3DXMESHOPT_ATTRSORT, adj,
            &adj_out[0], &face_remap[0], &vertex_remap));
    }
    // clone any per-vertex colors needed
    clone_teapot(m_clone_fvf);

    HMENU menu = ::GetMenu(m_hWnd);
    // can't select arbitrary color sources?
    bool enable =
        (m_d3dCaps.VertexProcessingCaps & D3DVTXPCAPS_MATERIALSOURCE7) != 0;
    ::enable_menu(menu, IDM_MATERIAL_AMBIENT_COLOR1, enable);
    ::enable_menu(menu, IDM_MATERIAL_AMBIENT_COLOR2, enable);
    ::enable_menu(menu, IDM_MATERIAL_DIFFUSE_COLOR1, enable);
    ::enable_menu(menu, IDM_MATERIAL_SPECULAR_COLOR2, enable);
    ::enable_menu(menu, IDM_MATERIAL_EMISSIVE_COLOR1, enable);
    ::enable_menu(menu, IDM_MATERIAL_EMISSIVE_COLOR2, enable);
    if (!enable)
    {
        if ((D3DMCS_COLOR1 == m_ambient_source) ||
            (D3DMCS_COLOR2 == m_ambient_source))
        {
            rt::check_menu(menu, IDM_MATERIAL_AMBIENT_COLOR1, false);
            rt::check_menu(menu, IDM_MATERIAL_AMBIENT_COLOR2, false);
            rt::check_menu(menu, IDM_MATERIAL_AMBIENT_MATERIAL, true);
            m_ambient_source = D3DMCS_MATERIAL;
        }
        if (D3DMCS_COLOR2 == m_diffuse_source)
        {
            rt::check_menu(menu, IDM_MATERIAL_DIFFUSE_COLOR2, false);
            rt::check_menu(menu, IDM_MATERIAL_DIFFUSE_MATERIAL, true);
            m_diffuse_source = D3DMCS_MATERIAL;
        }
        if (D3DMCS_COLOR1 == m_specular_source)
        {
            rt::check_menu(menu, IDM_MATERIAL_SPECULAR_COLOR1, false);
            rt::check_menu(menu, IDM_MATERIAL_SPECULAR_MATERIAL, true);
            m_specular_source = D3DMCS_MATERIAL;
        }
        if ((D3DMCS_COLOR1 == m_emissive_source) ||
            (D3DMCS_COLOR2 == m_emissive_source))
        {
            rt::check_menu(menu, IDM_MATERIAL_EMISSIVE_COLOR1, false);
            rt::check_menu(menu, IDM_MATERIAL_EMISSIVE_COLOR2, false);
            m_emissive_source = D3DMCS_MATERIAL;
        }
    }

    // can't perform gouraud shading?
    enable = (m_d3dCaps.ShadeCaps & D3DPSHADECAPS_COLORGOURAUDRGB) != 0;
    ::enable_menu(menu, IDM_SHADING_GOURAUD, enable);
    if (!enable && (D3DSHADE_GOURAUD == m_shade_mode))
    {
        rt::check_menu(menu, IDM_SHADING_FLAT, true);
        rt::check_menu(menu, IDM_SHADING_GOURAUD, false);
        m_shade_mode = D3DSHADE_FLAT;
    }

    // can't perform local viewer calculations?
    enable = (m_d3dCaps.VertexProcessingCaps & D3DVTXPCAPS_LOCALVIEWER) != 0;
    ::enable_menu(menu, IDM_OPTION_LOCAL_VIEWER, enable);
    m_local_viewer = enable && m_local_viewer;
    rt::check_menu(menu, IDM_OPTION_LOCAL_VIEWER, m_local_viewer);

    // can't handle all light types?
    switch (m_d3dCaps.VertexProcessingCaps &
            (D3DVTXPCAPS_DIRECTIONALLIGHTS | D3DVTXPCAPS_POSITIONALLIGHTS))
    {
    case D3DVTXPCAPS_DIRECTIONALLIGHTS | D3DVTXPCAPS_POSITIONALLIGHTS:
        // can do all light types
        break;

    case D3DVTXPCAPS_POSITIONALLIGHTS:
        // can't do directional lights
        ::enable_menu(menu, IDM_LIGHT_TYPE_DIRECTIONAL, false);
        if (D3DLIGHT_DIRECTIONAL == m_light.Type)
        {
            m_light.Type = D3DLIGHT_SPOT;
            update_light_menu(menu);
        }
        break;

    case D3DVTXPCAPS_DIRECTIONALLIGHTS:
        // can't do positional (point or spot) lights
        ::enable_menu(menu, IDM_LIGHT_TYPE_POINT, false);
        ::enable_menu(menu, IDM_LIGHT_TYPE_SPOT, false);
        if ((D3DLIGHT_POINT == m_light.Type) ||
            (D3DLIGHT_SPOT == m_light.Type))
        {
            m_light.Type = D3DLIGHT_DIRECTIONAL;
            update_light_menu(menu);
        }
        break;

    case 0:
        // can't do lights!
        ::enable_menu(menu, IDM_LIGHT_TYPE_DIRECTIONAL, false);
        ::enable_menu(menu, IDM_LIGHT_TYPE_POINT, false);
        ::enable_menu(menu, IDM_LIGHT_TYPE_SPOT, false);
        ::enable_menu(menu, IDM_LIGHT_ENABLE, false);
        rt::check_menu(menu, IDM_LIGHT_TYPE_DIRECTIONAL, false);
        rt::check_menu(menu, IDM_LIGHT_TYPE_POINT, false);
        rt::check_menu(menu, IDM_LIGHT_TYPE_SPOT, false);
        rt::check_menu(menu, IDM_LIGHT_ENABLE, false);
        m_light_enabled = false;
        break;

    default:
        ATLASSERT(false);
    }

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
    // Update user input state
    UpdateInput( &m_UserInput );

    // Update the world state according to user input
    D3DXMATRIX matWorld;
    D3DXMATRIX matRotY;
    D3DXMATRIX matRotX;

    // Update the world state according to user input
    if (m_animate_view)
    {
        m_fWorldRotY += m_fElapsedTime;
        if (m_fWorldRotY > 2*D3DX_PI)
        {
            m_fWorldRotY = fmodf(m_fWorldRotY, 2*D3DX_PI);
        }
    }
    else
    {
        if (m_UserInput.bRotateLeft && !m_UserInput.bRotateRight)
        {
            m_fWorldRotY += m_fElapsedTime;
        }
        else if (m_UserInput.bRotateRight && !m_UserInput.bRotateLeft)
        {
            m_fWorldRotY -= m_fElapsedTime;
        }
    }
    if (m_UserInput.bRotateUp && !m_UserInput.bRotateDown)
    {
        m_fWorldRotX += m_fElapsedTime;
    }
    else if (m_UserInput.bRotateDown && !m_UserInput.bRotateUp)
    {
        m_fWorldRotX -= m_fElapsedTime;
    }

    D3DXMatrixRotationX( &matRotX, m_fWorldRotX );
    D3DXMatrixRotationY( &matRotY, m_fWorldRotY );

    D3DXMatrixMultiply( &matWorld, &matRotX, &matRotY );
    m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: UpdateInput()
// Desc: Update the user input.  Called once per frame 
//-----------------------------------------------------------------------------
void CMyD3DApplication::UpdateInput( UserInput* pUserInput )
{
    pUserInput->bRotateUp    = ( m_bHasFocus && (GetAsyncKeyState( VK_UP )    & 0x8000) == 0x8000 );
    pUserInput->bRotateDown  = ( m_bHasFocus && (GetAsyncKeyState( VK_DOWN )  & 0x8000) == 0x8000 );
    pUserInput->bRotateLeft  = ( m_bHasFocus && (GetAsyncKeyState( VK_LEFT )  & 0x8000) == 0x8000 );
    pUserInput->bRotateRight = ( m_bHasFocus && (GetAsyncKeyState( VK_RIGHT ) & 0x8000) == 0x8000 );
}


void
join(rt::tostringstream &buff, bool val, LPCTSTR str)
{
    if (val)
    {
        if (buff.str().length())
        {
            buff << _T(", ");
        }
        buff << str;
    }
}


rt::tostream &
operator<<(rt::tostream &stream, D3DMATERIALCOLORSOURCE mcs)
{
    switch (mcs)
    {
    case D3DMCS_COLOR1:
        stream << _T("color1");
        break;

    case D3DMCS_COLOR2:
        stream << _T("color2");
        break;

    case D3DMCS_MATERIAL:
        stream << _T("material");
        break;

    default:
        ATLASSERT(false);
    }
    return stream;
}

rt::tostream &
operator<<(rt::tostream &stream, const D3DVECTOR &pos)
{
    return stream << _T("xyz(") << pos.x << _T(", ") << pos.y << _T(", ")
        << pos.z << _T(")");
}

rt::tostream &
operator<<(rt::tostream &stream, const D3DCOLORVALUE &colorval)
{
    return stream << _T("rgba(") << colorval.r << _T(", ") << colorval.g
        << _T(", ") << colorval.b << _T(", ") << colorval.a << _T(")");
}

//-----------------------------------------------------------------------------
// Name: RenderText()
// Desc: Renders stats and help text to the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RenderText()
{
    D3DCOLOR fontColor        = D3DCOLOR_ARGB(255,255,255,0);
    D3DCOLOR fontWarningColor = D3DCOLOR_ARGB(255,0,255,255);
    TCHAR szMsg[MAX_PATH] = TEXT("");

    // Output display stats
    FLOAT fNextLine = 40.0f; 

    lstrcpy( szMsg, m_strDeviceStats );
    fNextLine -= 20.0f;
    m_pFont->DrawText( 2, fNextLine, fontColor, szMsg );

    lstrcpy( szMsg, m_strFrameStats );
    fNextLine -= 20.0f;
    m_pFont->DrawText( 2, fNextLine, fontColor, szMsg );

    fNextLine = 40.0f;
    {
        rt::tostringstream buff;
        join(buff, m_lighting, _T("lighting"));
        join(buff, m_local_viewer, _T("local viewer"));
        join(buff, m_normalize_normals, _T("normalize normals"));
        join(buff, m_specular_enable, _T("specular"));
        join(buff, m_dithering, _T("dithering"));
        m_pFont->DrawText(2, fNextLine, fontColor, const_cast<LPTSTR>(buff.str().c_str()));
    }
    fNextLine += 20.0f;
#if 0
    {
        rt::tostringstream buff;
        buff << _T("Ambient: ") << m_ambient_source;
        if (D3DMCS_MATERIAL == m_ambient_source)
        {
            buff << _T(", material: ") << m_material.Ambient;
        }
        buff << _T(", light: ") << m_light.Ambient;
        m_pFont->DrawText(2, fNextLine, fontColor, const_cast<LPTSTR>(buff.str().c_str()));
    }
    fNextLine += 20.0f;
    {
        rt::tostringstream buff;
        buff << _T("Diffuse: ") << m_diffuse_source << _T(", material: ")
            << m_material.Diffuse << _T(", light: ")
            << m_light.Diffuse;
        m_pFont->DrawText(2, fNextLine, fontColor, const_cast<LPTSTR>(buff.str().c_str()));
    }
    fNextLine += 20.0f;
    {
        rt::tostringstream buff;
        buff << _T("Specular: ") << m_specular_source << _T(", material: ")
            << m_material.Specular << _T(", light: ")
            << m_light.Specular;
        m_pFont->DrawText(2, fNextLine, fontColor, const_cast<LPTSTR>(buff.str().c_str()));
    }
    fNextLine += 20.0f;
    {
        rt::tostringstream buff;
        buff << _T("Emissive: ") << m_emissive_source << _T(", material: ")
            << m_material.Emissive;
        m_pFont->DrawText(2, fNextLine, fontColor, const_cast<LPTSTR>(buff.str().c_str()));
    }
#endif

    // Output statistics & help
    fNextLine = (FLOAT) m_d3dsdBackBuffer.Height; 

    wsprintf( szMsg, TEXT("Arrow keys: Up=%d Down=%d Left=%d Right=%d"), 
              m_UserInput.bRotateUp, m_UserInput.bRotateDown, m_UserInput.bRotateLeft, m_UserInput.bRotateRight );
    fNextLine -= 20.0f; m_pFont->DrawText( 2, fNextLine, fontColor, szMsg );

    sprintf( szMsg, TEXT("World State: %0.3f, %0.3f"), 
                    m_fWorldRotX, m_fWorldRotY );
    fNextLine -= 20.0f; m_pFont->DrawText( 2, fNextLine, fontColor, szMsg );

    lstrcpy( szMsg, TEXT("Use arrow keys to update input") );
    fNextLine -= 20.0f; m_pFont->DrawText( 2, fNextLine, fontColor, szMsg );

    lstrcpy( szMsg, TEXT("Press 'F2' to configure display") );
    fNextLine -= 20.0f; m_pFont->DrawText( 2, fNextLine, fontColor, szMsg );

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Overrrides the main WndProc, so the sample can do custom message
//       handling (e.g. processing mouse, keyboard, or menu commands).
//-----------------------------------------------------------------------------
LRESULT CMyD3DApplication::MsgProc( HWND hWnd, UINT msg, WPARAM wParam,
                                    LPARAM lParam )
{
    switch( msg )
    {
    case WM_PAINT:
        if( m_bLoadingApp )
        {
            // Draw on the window tell the user that the app is loading
            HDC hDC = GetDC( hWnd );
            TCHAR strMsg[MAX_PATH];
            wsprintf( strMsg, TEXT("Loading... Please wait") );
            RECT rct;
            GetClientRect( hWnd, &rct );
            DrawText( hDC, strMsg, -1, &rct, DT_CENTER|DT_VCENTER|DT_SINGLELINE );
            ReleaseDC( hWnd, hDC );
        }
        break;

    case WM_COMMAND:
        on_command(hWnd, wParam, lParam);
        break;
    }

    return CD3DApplication::MsgProc( hWnd, msg, wParam, lParam );
}




//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Invalidates device objects.  
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
    m_pFont->InvalidateDeviceObjects();

    m_teapot_nocolors = 0;
    m_teapot = 0;

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.  
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
    m_pFont->DeleteDeviceObjects();

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
    // Cleanup D3D font
    SAFE_DELETE( m_pFont );

    // Write the settings to the registry
    WriteSettings();

    return S_OK;
}



//////////////////////////////////////////////////////////////////////
// CMyD3DApplication::on_command
//
// WM_COMMAND message handler
//
void
CMyD3DApplication::on_command(HWND window, WPARAM wp, LPARAM lp)
{
    const HMENU menu = ::GetMenu(window);
    const UINT control = LOWORD(wp);

    switch (control)
    {
    case IDM_VERTEX_COLOR_NONE:
        clone_teapot(0);
        rt::check_menu(menu, IDM_VERTEX_COLOR_NONE, true);
        rt::check_menu(menu, IDM_VERTEX_COLOR_DIFFUSE, false);
        rt::check_menu(menu, IDM_VERTEX_COLOR_SPECULAR, false);
        rt::check_menu(menu, IDM_VERTEX_COLOR_DIFFUSE_SPECULAR, false);
        break;

    case IDM_VERTEX_COLOR_DIFFUSE:
        clone_teapot(D3DFVF_DIFFUSE);
        rt::check_menu(menu, IDM_VERTEX_COLOR_NONE, false);
        rt::check_menu(menu, IDM_VERTEX_COLOR_DIFFUSE, true);
        rt::check_menu(menu, IDM_VERTEX_COLOR_SPECULAR, false);
        rt::check_menu(menu, IDM_VERTEX_COLOR_DIFFUSE_SPECULAR, false);
        break;

    case IDM_VERTEX_COLOR_SPECULAR:
        clone_teapot(D3DFVF_SPECULAR);
        rt::check_menu(menu, IDM_VERTEX_COLOR_NONE, false);
        rt::check_menu(menu, IDM_VERTEX_COLOR_DIFFUSE, false);
        rt::check_menu(menu, IDM_VERTEX_COLOR_SPECULAR, true);
        rt::check_menu(menu, IDM_VERTEX_COLOR_DIFFUSE_SPECULAR, false);
        break;

    case IDM_VERTEX_COLOR_DIFFUSE_SPECULAR:
        clone_teapot(D3DFVF_DIFFUSE | D3DFVF_SPECULAR);
        rt::check_menu(menu, IDM_VERTEX_COLOR_NONE, false);
        rt::check_menu(menu, IDM_VERTEX_COLOR_DIFFUSE, false);
        rt::check_menu(menu, IDM_VERTEX_COLOR_SPECULAR, false);
        rt::check_menu(menu, IDM_VERTEX_COLOR_DIFFUSE_SPECULAR, true);
        break;

#define MODIFY_MATERIAL_COLOR(id_, member_, transparent_)                   \
    case id_:                                                               \
        {                                                                   \
            rt::pauser block(*this);                                        \
            m_material.member_ = D3DXCOLOR(choose_color_transparent(window, \
                D3DXCOLOR(m_material.member_), transparent_));              \
        }                                                                   \
        break
    MODIFY_MATERIAL_COLOR(IDM_MATERIAL_DIFFUSE, Diffuse, true);
    MODIFY_MATERIAL_COLOR(IDM_MATERIAL_SPECULAR, Specular, false);
    MODIFY_MATERIAL_COLOR(IDM_MATERIAL_AMBIENT, Ambient, false);
    MODIFY_MATERIAL_COLOR(IDM_MATERIAL_EMISSIVE, Emissive, true);
#undef MODIFY_MATERIAL_COLOR

#define MODIFY_MATERIAL_SOURCE(state_, id_)     \
    case IDM_MATERIAL_##id_##_MATERIAL:         \
    case IDM_MATERIAL_##id_##_COLOR1:           \
    case IDM_MATERIAL_##id_##_COLOR2:           \
        rt::check_menu(menu, IDM_MATERIAL_##id_##_MATERIAL + UINT(state_), false); \
        state_ = D3DMATERIALCOLORSOURCE(control - IDM_MATERIAL_##id_##_MATERIAL); \
        rt::check_menu(menu, IDM_MATERIAL_##id_##_MATERIAL + UINT(state_), true); \
        break
    MODIFY_MATERIAL_SOURCE(m_ambient_source, AMBIENT);
    MODIFY_MATERIAL_SOURCE(m_diffuse_source, DIFFUSE);
    MODIFY_MATERIAL_SOURCE(m_specular_source, SPECULAR);
    MODIFY_MATERIAL_SOURCE(m_emissive_source, EMISSIVE);
#undef MODIFY_MATERIAL_SOURCE

    case IDM_MATERIAL_SPECULAR_POWER:
        update_material_parameters(IDD_MATERIAL_SPECULAR_POWER, material_specular_power_proc);
        break;

#define MODIFY_LIGHT_REFLECTANCE(id_, member_, transparent_)                \
    case id_:                                                               \
        {                                                                   \
            rt::pauser block(*this);                                        \
            m_light.member_ = D3DXCOLOR(choose_color_transparent(window,    \
                D3DXCOLOR(m_light.member_), transparent_));                 \
        }                                                                   \
        break
    MODIFY_LIGHT_REFLECTANCE(IDM_LIGHT_AMBIENT, Ambient, false);
    MODIFY_LIGHT_REFLECTANCE(IDM_LIGHT_DIFFUSE, Diffuse, true);
    MODIFY_LIGHT_REFLECTANCE(IDM_LIGHT_SPECULAR, Specular, false);
#undef MODIFY_LIGHT_REFLECTANCE

    case IDM_LIGHT_TYPE_POINT:
    case IDM_LIGHT_TYPE_SPOT:
    case IDM_LIGHT_TYPE_DIRECTIONAL:
        m_light.Type = D3DLIGHTTYPE(control - IDM_LIGHT_TYPE_POINT + 1);
        update_light_menu(menu);
        break;

    case IDM_LIGHT_POSITION:
        update_light_parameters(IDD_LIGHT_POSITION, light_position_proc);
        break;

    case IDM_LIGHT_DIRECTION:
        update_light_parameters(IDD_LIGHT_DIRECTION, light_direction_proc);
        break;

    case IDM_LIGHT_RANGE:
        update_light_parameters(IDD_LIGHT_RANGE, light_range_proc);
        break;

    case IDM_LIGHT_ATTENUATION:
        update_light_parameters(IDD_LIGHT_ATTENUATION, light_attenuation_proc);
        break;

    case IDM_LIGHT_CONE:
        update_light_parameters(IDD_LIGHT_CONE, light_cone_proc);
        break;

    case IDM_SHADING_FLAT:
        m_shade_mode = D3DSHADE_FLAT;
        rt::check_menu(menu, IDM_SHADING_FLAT, true);
        rt::check_menu(menu, IDM_SHADING_GOURAUD, false);
        break;

    case IDM_SHADING_GOURAUD:
        m_shade_mode = D3DSHADE_GOURAUD;
        rt::check_menu(menu, IDM_SHADING_FLAT, false);
        rt::check_menu(menu, IDM_SHADING_GOURAUD, true);
        break;

    case IDM_FILL_POINT:
        m_fill_mode = D3DFILL_POINT;
        rt::check_menu(menu, IDM_FILL_POINT, true);
        rt::check_menu(menu, IDM_FILL_WIREFRAME, false);
        rt::check_menu(menu, IDM_FILL_SOLID, false);
        break;

    case IDM_FILL_WIREFRAME:
        m_fill_mode = D3DFILL_WIREFRAME;
        rt::check_menu(menu, IDM_FILL_POINT, false);
        rt::check_menu(menu, IDM_FILL_WIREFRAME, true);
        rt::check_menu(menu, IDM_FILL_SOLID, false);
        break;

    case IDM_FILL_SOLID:
        m_fill_mode = D3DFILL_SOLID;
        rt::check_menu(menu, IDM_FILL_POINT, false);
        rt::check_menu(menu, IDM_FILL_WIREFRAME, false);
        rt::check_menu(menu, IDM_FILL_SOLID, true);
        break;

    case IDM_PROJECTION_ORTHOGRAPHIC:
        rt::check_menu(menu, IDM_PROJECTION_PERSPECTIVE, false);
        rt::check_menu(menu, IDM_PROJECTION_ORTHOGRAPHIC, true);
        m_projection = PT_ORTHOGRAPHIC;
        set_projection();
        break;

    case IDM_PROJECTION_PERSPECTIVE:
        rt::check_menu(menu, IDM_PROJECTION_PERSPECTIVE, true);
        rt::check_menu(menu, IDM_PROJECTION_ORTHOGRAPHIC, false);
        m_projection = PT_PERSPECTIVE;
        set_projection();
        break;

#define TOGGLE(id_, state_)                 \
    case id_:                               \
        state_ = !state_;                   \
        rt::check_menu(menu, id_, state_);  \
        break
    TOGGLE(IDM_LIGHT_ENABLE, m_light_enabled);
    TOGGLE(IDM_OPTION_ANIMATE_VIEW, m_animate_view);
    TOGGLE(IDM_OPTION_SHOW_TEXT, m_show_text);
    TOGGLE(IDM_OPTION_NORMALIZE_NORMALS, m_normalize_normals);
    TOGGLE(IDM_OPTION_LOCAL_VIEWER, m_local_viewer);
    TOGGLE(IDM_OPTION_LIGHTING, m_lighting);
    TOGGLE(IDM_OPTION_SPECULAR_ENABLE, m_specular_enable);
    TOGGLE(IDM_OPTION_DITHER, m_dithering);
    TOGGLE(IDM_OPTION_COLOR_VERTEX, m_color_vertex);
#undef TOGGLE

    case IDM_OPTION_AMBIENT_COLOR:
        {
            rt::pauser block(*this);
            m_ambient = choose_color(window, m_ambient);
        }
        break;

    case IDM_OPTION_RESET_VIEW:
        m_fWorldRotX = 0.0f;
        m_fWorldRotY = 0.0f;
        break;

    case IDM_OPTION_BACKGROUND_COLOR:
        {
            rt::pauser block(*this);
            m_background = choose_color_transparent(window, m_background, true);
        }
        break;

    default:
        // all our control IDs are > 40006 and we should handle them all
        if (control > 40006)
        {
            ATLASSERT(false);
        }
    }
}

//////////////////////////////////////////////////////////////////////
// CMyD3DApplication::update_light_menu
//
// When changing the light type, we need to enable or disable the
// relevant lighting parameter menu items.
//
void
CMyD3DApplication::update_light_menu(HMENU menu)
{
    switch (m_light.Type)
    {
    case D3DLIGHT_POINT:
        rt::check_menu(menu, IDM_LIGHT_TYPE_POINT, true);
        rt::check_menu(menu, IDM_LIGHT_TYPE_SPOT, false);
        rt::check_menu(menu, IDM_LIGHT_TYPE_DIRECTIONAL, false);
        ::enable_menu(menu, IDM_LIGHT_POSITION, true);
        ::enable_menu(menu, IDM_LIGHT_DIRECTION, false);
        ::enable_menu(menu, IDM_LIGHT_RANGE, true);
        ::enable_menu(menu, IDM_LIGHT_ATTENUATION, true);
        ::enable_menu(menu, IDM_LIGHT_CONE, false);
        break;

    case D3DLIGHT_SPOT:
        rt::check_menu(menu, IDM_LIGHT_TYPE_POINT, false);
        rt::check_menu(menu, IDM_LIGHT_TYPE_SPOT, true);
        rt::check_menu(menu, IDM_LIGHT_TYPE_DIRECTIONAL, false);
        ::enable_menu(menu, IDM_LIGHT_POSITION, true);
        ::enable_menu(menu, IDM_LIGHT_DIRECTION, true);
        ::enable_menu(menu, IDM_LIGHT_RANGE, true);
        ::enable_menu(menu, IDM_LIGHT_ATTENUATION, true);
        ::enable_menu(menu, IDM_LIGHT_CONE, true);
        break;

    case D3DLIGHT_DIRECTIONAL:
        rt::check_menu(menu, IDM_LIGHT_TYPE_POINT, false);
        rt::check_menu(menu, IDM_LIGHT_TYPE_SPOT, false);
        rt::check_menu(menu, IDM_LIGHT_TYPE_DIRECTIONAL, true);
        ::enable_menu(menu, IDM_LIGHT_POSITION, false);
        ::enable_menu(menu, IDM_LIGHT_DIRECTION, true);
        ::enable_menu(menu, IDM_LIGHT_RANGE, false);
        ::enable_menu(menu, IDM_LIGHT_ATTENUATION, false);
        ::enable_menu(menu, IDM_LIGHT_CONE, false);
        break;

    default:
        ATLASSERT(false);
    }
}

//////////////////////////////////////////////////////////////////////
// CMyD3DApplication::update_material_parameters
//
// Invoke a dialog box to edit some material parameters.  The dialog
// box edits a temporary copy of the application's material parameters
// and if the dialog box is ended with the OK button, then we copy
// the edited parameters into the application.
//
void
CMyD3DApplication::update_material_parameters(UINT dialog_id, DialogProc *proc)
{
    rt::pauser block(*this);

    D3DMATERIAL8 material = m_material;
    if (IDOK == DialogBoxParam(HINSTANCE(GetModuleHandle(NULL)),
                               MAKEINTRESOURCE(dialog_id), m_hWnd,
                               proc, reinterpret_cast<LPARAM>(&material)))
    {
        m_material = material;
    }
}

//////////////////////////////////////////////////////////////////////
// update_float
//
// Helper routine for managing the UI state of edit boxes tied to
// floating-point parameters.
//
void
update_float(HWND dialog, UINT control_id, float &val, bool get_not_put)
{
    HWND edit = TWS(::GetDlgItem(dialog, control_id));
    if (get_not_put)
    {
        rt::tostringstream buff;
        buff << val << std::ends;
        Edit_SetSel(edit, 0, -1);
        Edit_ReplaceSel(edit, buff.str().c_str());
    }
    else
    {
        std::vector<TCHAR> edit_text(20);
        {
            const WORD text_len = WORD(Edit_LineLength(edit, 0));
            edit_text.resize(text_len);

            if (text_len > 0)
            {
                ::memcpy(&edit_text[0], &text_len, sizeof(WORD));
                ::SendMessage(edit, EM_GETLINE, 0, LPARAM(&edit_text[0]));
                edit_text[text_len] = 0;
            }
        }

        rt::tistringstream stream(&edit_text[0]);
        float value;
        stream >> value;
        if (!stream.fail())
        {
            val = value;
        }
    }
}

//////////////////////////////////////////////////////////////////////
// CMyD3DApplication::material_specular_power_proc
//
// Dialog box window procedure for editing the specular power material
// parameter.
//
INT_PTR CALLBACK
CMyD3DApplication::material_specular_power_proc(HWND dialog, UINT msg,
                                                WPARAM wp, LPARAM lp)
{
    BOOL result = FALSE;

    // Static state for adapter/device/mode selection
    static D3DMATERIAL8 *material = NULL;

    // Handle the initialization message
    switch (msg)
    {
    case WM_INITDIALOG:
        material = reinterpret_cast<D3DMATERIAL8 *>(lp);

        // Set flag to update dialog controls below
        ::update_float(dialog, IDC_EDIT_SPECULAR_POWER, material->Power, true);
        result = TRUE;
        break;

    case WM_COMMAND:
        switch (LOWORD(wp))
        {
        case IDOK:
            ::update_float(dialog, IDC_EDIT_SPECULAR_POWER, material->Power, false);
            ::EndDialog(dialog, IDOK);
            result = TRUE;
            break;

        case IDCANCEL:
            ::EndDialog(dialog, IDCANCEL);
            result = TRUE;
            break;
        } // switch
        break;
    }

    return result;
}

//////////////////////////////////////////////////////////////////////
// CMyD3DApplication::update_light_parameters
//
// Invoke a dialog box to edit some light parameters.  The dialog
// box edits a temporary copy of the application's light parameters
// and if the dialog box is ended with the OK button, then we copy
// the edited parameters into the application.
//
void
CMyD3DApplication::update_light_parameters(UINT dialog_id, DialogProc *proc)
{
    rt::pauser block(*this);

    D3DLIGHT8 light = m_light;
    if (IDOK == DialogBoxParam(HINSTANCE(GetModuleHandle(NULL)),
                               MAKEINTRESOURCE(dialog_id), m_hWnd,
                               proc, reinterpret_cast<LPARAM>(&light)))
    {
        m_light = light;
    }
}

//////////////////////////////////////////////////////////////////////
// CMyD3DApplication::light_direction_proc
//
// Dialog box window procedure for editing the light direction parameter.
//
INT_PTR CALLBACK
CMyD3DApplication::light_direction_proc(HWND dialog, UINT msg,
                                        WPARAM wp, LPARAM lp)
{
    BOOL result = FALSE;

    // Static state for adapter/device/mode selection
    static D3DLIGHT8 *light = NULL;

    // Handle the initialization message
    switch (msg)
    {
    case WM_INITDIALOG:
        light = reinterpret_cast<D3DLIGHT8 *>(lp);

        // Set flag to update dialog controls below
        ::update_float(dialog, IDC_EDIT_DIRECTION_X, light->Direction.x, true);
        ::update_float(dialog, IDC_EDIT_DIRECTION_Y, light->Direction.y, true);
        ::update_float(dialog, IDC_EDIT_DIRECTION_Z, light->Direction.z, true);
        result = TRUE;
        break;

    case WM_COMMAND:
        switch (LOWORD(wp))
        {
        case IDOK:
            ::update_float(dialog, IDC_EDIT_DIRECTION_X, light->Direction.x, false);
            ::update_float(dialog, IDC_EDIT_DIRECTION_Y, light->Direction.y, false);
            ::update_float(dialog, IDC_EDIT_DIRECTION_Z, light->Direction.z, false);
            ::EndDialog(dialog, IDOK);
            result = TRUE;
            break;

        case IDCANCEL:
            ::EndDialog(dialog, IDCANCEL);
            result = TRUE;
            break;
        } // switch
        break;
    }

    return result;
}

//////////////////////////////////////////////////////////////////////
// CMyD3DApplication::light_position_proc
//
// Dialog box window procedure for editing the light position parameter.
//
INT_PTR CALLBACK
CMyD3DApplication::light_position_proc(HWND dialog, UINT msg,
                                       WPARAM wp, LPARAM lp)
{
    BOOL result = FALSE;

    // Static state for adapter/device/mode selection
    static D3DLIGHT8 *light = NULL;

    // Handle the initialization message
    switch (msg)
    {
    case WM_INITDIALOG:
        light = reinterpret_cast<D3DLIGHT8 *>(lp);

        // Set flag to update dialog controls below
        ::update_float(dialog, IDC_EDIT_POSITION_X, light->Position.x, true);
        ::update_float(dialog, IDC_EDIT_POSITION_Y, light->Position.y, true);
        ::update_float(dialog, IDC_EDIT_POSITION_Z, light->Position.z, true);
        result = TRUE;
        break;

    case WM_COMMAND:
        switch (LOWORD(wp))
        {
        case IDOK:
            ::update_float(dialog, IDC_EDIT_POSITION_X, light->Position.x, false);
            ::update_float(dialog, IDC_EDIT_POSITION_Y, light->Position.y, false);
            ::update_float(dialog, IDC_EDIT_POSITION_Z, light->Position.z, false);
            ::EndDialog(dialog, IDOK);
            result = TRUE;
            break;

        case IDCANCEL:
            ::EndDialog(dialog, IDCANCEL);
            result = TRUE;
            break;
        } // switch
        break;
    }

    return result;
}

//////////////////////////////////////////////////////////////////////
// CMyD3DApplication::light_range_proc
//
// Dialog box window procedure for editing the light range and falloff
// parameters.
//
INT_PTR CALLBACK
CMyD3DApplication::light_range_proc(HWND dialog, UINT msg,
                                    WPARAM wp, LPARAM lp)
{
    BOOL result = FALSE;

    // Static state for adapter/device/mode selection
    static D3DLIGHT8 *light = NULL;

    // Handle the initialization message
    switch (msg)
    {
    case WM_INITDIALOG:
        light = reinterpret_cast<D3DLIGHT8 *>(lp);

        // Set flag to update dialog controls below
        ::update_float(dialog, IDC_EDIT_RANGE, light->Range, true);
        ::update_float(dialog, IDC_EDIT_FALLOFF, light->Falloff, true);
        result = TRUE;
        break;

    case WM_COMMAND:
        switch (LOWORD(wp))
        {
        case IDOK:
            ::update_float(dialog, IDC_EDIT_RANGE, light->Range, false);
            ::update_float(dialog, IDC_EDIT_FALLOFF, light->Falloff, false);
            ::EndDialog(dialog, IDOK);
            result = TRUE;
            break;

        case IDCANCEL:
            ::EndDialog(dialog, IDCANCEL);
            result = TRUE;
            break;
        } // switch
        break;
    }

    return result;
}

//////////////////////////////////////////////////////////////////////
// CMyD3DApplication::light_attenuation_proc
//
// Dialog box window procedure for editing the light attenuation
// parameters.
//
INT_PTR CALLBACK
CMyD3DApplication::light_attenuation_proc(HWND dialog, UINT msg,
                                          WPARAM wp, LPARAM lp)
{
    BOOL result = FALSE;

    // Static state for adapter/device/mode selection
    static D3DLIGHT8 *light = NULL;

    // Handle the initialization message
    switch (msg)
    {
    case WM_INITDIALOG:
        light = reinterpret_cast<D3DLIGHT8 *>(lp);

        // Set flag to update dialog controls below
        ::update_float(dialog, IDC_EDIT_ATTENUATION_0, light->Attenuation0, true);
        ::update_float(dialog, IDC_EDIT_ATTENUATION_1, light->Attenuation1, true);
        ::update_float(dialog, IDC_EDIT_ATTENUATION_2, light->Attenuation2, true);
        result = TRUE;
        break;

    case WM_COMMAND:
        switch (LOWORD(wp))
        {
        case IDOK:
            ::update_float(dialog, IDC_EDIT_ATTENUATION_0, light->Attenuation0, false);
            ::update_float(dialog, IDC_EDIT_ATTENUATION_1, light->Attenuation1, false);
            ::update_float(dialog, IDC_EDIT_ATTENUATION_2, light->Attenuation2, false);
            ::EndDialog(dialog, IDOK);
            result = TRUE;
            break;

        case IDCANCEL:
            ::EndDialog(dialog, IDCANCEL);
            result = TRUE;
            break;
        } // switch
        break;
    }

    return result;
}

//////////////////////////////////////////////////////////////////////
// CMyD3DApplication::light_cone_proc
//
// Dialog box window procedure for editing the light cone angle
// parameters.
//
INT_PTR CALLBACK
CMyD3DApplication::light_cone_proc(HWND dialog, UINT msg,
                                   WPARAM wp, LPARAM lp)
{
    BOOL result = FALSE;

    // Static state for adapter/device/mode selection
    static D3DLIGHT8 *light = NULL;

    // Handle the initialization message
    switch (msg)
    {
    case WM_INITDIALOG:
        light = reinterpret_cast<D3DLIGHT8 *>(lp);

        // Set flag to update dialog controls below
        ::update_float(dialog, IDC_EDIT_THETA, light->Theta, true);
        ::update_float(dialog, IDC_EDIT_PHI, light->Phi, true);
        result = TRUE;
        break;

    case WM_COMMAND:
        switch (LOWORD(wp))
        {
        case IDOK:
            ::update_float(dialog, IDC_EDIT_THETA, light->Theta, false);
            ::update_float(dialog, IDC_EDIT_PHI, light->Phi, false);
            ::EndDialog(dialog, IDOK);
            result = TRUE;
            break;

        case IDCANCEL:
            ::EndDialog(dialog, IDCANCEL);
            result = TRUE;
            break;
        } // switch
        break;
    }

    return result;
}

//////////////////////////////////////////////////////////////////////
// hsv_to_rgb
//
// Convert hue, saturation, value to red, green, blue.
//
void
hsv_to_rgb(float h, float s, float v,
           float &r, float &g, float &b)
{
    float f, p1, p2, p3;

    if (s < 0.0f || s > 1.0f)
      {
        rt::tostringstream msg;
        msg << _T("hsv_to_rgb: saturation parameter out of range: ") << s << _T("\n");
        ::OutputDebugString(msg.str().c_str());
        return;
      }

    if (v < 0.0f || v > 1.0f)
      {
        rt::tostringstream msg;
        msg << _T("hsv_to_rgb: value parameter out of range: ") << v << _T("\n");
        ::OutputDebugString(msg.str().c_str());
        return;
      }

    if (h < 0.0f)
      h -= ((int) (h/360.0f)-1.0f)*360.0f;
    else if (h >= 360.0f)
      h -= ((int) (h/360.))*360.0f;

    h /= 60.0f;

    f = h - ((int) h);

    p1 = v * (1.0f - s);
    p2 = v * (1.0f - (s * f));
    p3 = v * (1.0f - (s * (1. - f)));

    switch ((int) h)
    {
        case 0: r = v,  g = p3, b = p1; break;
        case 1: r = p2, g = v,  b = p1; break;
        case 2: r = p1, g = v,  b = p3; break;
        case 3: r = p1, g = p2, b =  v; break;
        case 4: r = p3, g = p1, b =  v; break;
        case 5: r = v,  g = p1, b = p2; break;
    }
}

//////////////////////////////////////////////////////////////////////
// CMyD3DApplication::set_projection
//
// Set the projection matrix to either perspective or orthographic.
//
void
CMyD3DApplication::set_projection()
{
    const float aspect =
        float(m_d3dsdBackBuffer.Width)/m_d3dsdBackBuffer.Height;
    D3DXMATRIX proj;
    if (PT_PERSPECTIVE == m_projection)
    {
        ::D3DXMatrixPerspectiveFovLH(&proj, D3DX_PI/4, aspect, 2.0f, 20.0f);
    }
    else
    {
        ::D3DXMatrixOrthoLH(&proj, 4*aspect, 4, 2.0f, 20.0f);
    }
    THR(m_pd3dDevice->SetTransform(D3DTS_PROJECTION, &proj));
}
