////////////////////////////////////////////////////////////
// rt_Lighting.cpp
//
// Interactive exploration of parameters affecting lights
// and materials.
//
// Copyright (C) 2002, Pahvant Technologies
// All rights reserved.
//
#define STRICT
#include <windows.h>

#include <d3dx8.h>

#include "D3DApp.h"
#include "D3DFont.h"

#include "rt_Lighting.h"

#include "rt/hr.h"
#include "rt/mat.h"
#include "rt/mesh.h"
#include "rt/misc.h"
#include "rt/states.h"

////////////////////////////////////////////////////////////
// vtx_teapot
//
// Vertex structure given back by ::D3DXCreateTeapot
//
struct vtx_teapot
{
    D3DXVECTOR3 position;
    D3DXVECTOR3 normal;
};

////////////////////////////////////////////////////////////
// vtx_diffuse
//
// A teapot vertex with a diffuse color.
//
struct vtx_diffuse : vtx_teapot
{
    D3DCOLOR diffuse;
};

////////////////////////////////////////////////////////////
// vtx_specular
//
// A teapot vertex with a specular color.
//
struct vtx_specular : vtx_teapot
{
    D3DCOLOR specular;
};

////////////////////////////////////////////////////////////
// vtx_diffuse_specular
//
// A teapot vertex with a diffuse and specular color.
//
struct vtx_diffuse_specular : vtx_teapot
{
    D3DCOLOR diffuse;
    D3DCOLOR specular;
};

////////////////////////////////////////////////////////////
// bounding_box<Vertex>
//
// Compute the bounding box for a mesh.
//
template <typename Vertex>
class bounding_box
{
public:
    bounding_box(ID3DXMesh *mesh)
        : m_minima(0, 0, 0),
        m_maxima(0, 0, 0)
    {
        rt::mesh_vertex_lock<Vertex> lock(mesh);
        THR(::D3DXComputeBoundingBox(lock.data(),
            mesh->GetNumVertices(), mesh->GetFVF(),
            &m_minima, &m_maxima));
    }
    ~bounding_box()
    {}

    const D3DXVECTOR3 &minima() const { return m_minima; }
    const D3DXVECTOR3 &maxima() const { return m_maxima; }

private:
    D3DXVECTOR3 m_minima;
    D3DXVECTOR3 m_maxima;
};

////////////////////////////////////////////////////////////
// diffuse_color
//
// Compute a diffuse color for the teapot by creating a hue
// ramp in the x direction.
//
D3DCOLOR
diffuse_color(const D3DXVECTOR3 &position)
{
    const float hue = (position.x + 1.0f)*360.0f;
    D3DCOLORVALUE diffuse = { 0 };
    hsv_to_rgb(hue, 1.0f, 1.0f,
        diffuse.r, diffuse.g, diffuse.b);
    diffuse.a = 1.0f;
    return D3DXCOLOR(diffuse);
}

////////////////////////////////////////////////////////////
// specular_color
//
// Compute a specular color by creating a grayscale gradient
// over the top half of the teapot.
//
D3DCOLOR
specular_color(const D3DXVECTOR3 &position,
               const D3DXVECTOR3 &minima,
               const D3DXVECTOR3 &maxima)
{
    const float fraction =
        (position.y - minima.y)/(maxima.y - minima.y);
    const BYTE intensity = fraction > 0.5f ?
        BYTE(255*min((fraction - 0.5f)*2.0f, 1.0f)) : 0;

    return D3DCOLOR_XRGB(intensity, intensity, intensity);
}

////////////////////////////////////////////////////////////
// CMyD3DApplication::clone_teapot
//
// Clone the teapot mesh into a mesh with the requested
// colors for each vertex.  Then compute the requested
// colors into the cloned data.
//
void
CMyD3DApplication::clone_teapot(DWORD fvf)
{
    // fvf must only contain specular, diffuse, or both
    ATLASSERT(!(fvf & ~(D3DFVF_DIFFUSE | D3DFVF_SPECULAR)));

    m_teapot = 0;
    m_clone_fvf = fvf;
    if (!fvf)
    {
        m_teapot = m_teapot_nocolors;
        return;
    }

    THR(m_teapot_nocolors->CloneMeshFVF(D3DXMESH_MANAGED,
        m_teapot_nocolors->GetFVF() | fvf, m_pd3dDevice,
        &m_teapot));

    if (D3DFVF_DIFFUSE == fvf)
    {
        rt::mesh_vertex_lock<vtx_diffuse> lock(m_teapot);
        vtx_diffuse *verts = lock.data();

        for (UINT i = 0; i < m_teapot->GetNumVertices(); i++)
        {
            verts[i].diffuse =
                diffuse_color(verts[i].position);
        }
    }
    else if (D3DFVF_SPECULAR == fvf)
    {
        rt::mesh_vertex_lock<vtx_specular> lock(m_teapot);
        vtx_specular *verts = lock.data();
        bounding_box<vtx_specular> minmax(m_teapot);

        for (UINT i = 0; i < m_teapot->GetNumVertices(); i++)
        {
            verts[i].specular =
                specular_color(verts[i].position,
                    minmax.minima(), minmax.maxima());
        }
    }
    else // ((D3DFVF_DIFFUSE | D3DFVF_SPECULAR) == fvf)
    {
        rt::mesh_vertex_lock<vtx_diffuse_specular>
            lock(m_teapot);
        vtx_diffuse_specular *verts = lock.data();
        bounding_box<vtx_diffuse_specular> minmax(m_teapot);

        for (UINT i = 0; i < m_teapot->GetNumVertices(); i++)
        {
            verts[i].diffuse =
                diffuse_color(verts[i].position);
            verts[i].specular =
                specular_color(verts[i].position,
                    minmax.minima(), minmax.maxima());
        }
    }
}

////////////////////////////////////////////////////////////
// CMyD3DApplication::Render
//
// Render a teapot with lighting and material parameters
// coming from the currently set material property on the
// device.
//
HRESULT
CMyD3DApplication::Render()
{
    // clear the frame buffer and start the scene
    THR(m_pd3dDevice->Clear(0L, NULL, D3DCLEAR_TARGET |
        D3DCLEAR_ZBUFFER, m_background, 1.0f, 0L));
    THR(m_pd3dDevice->BeginScene());

    // set the light and enable it, if necessary
    THR(m_pd3dDevice->SetLight(0, &m_light));
    THR(m_pd3dDevice->LightEnable(0, m_light_enabled));

    // set the material parameters
    THR(m_pd3dDevice->SetMaterial(&m_material));

    // set lighting-related and other render states
    rt::s_rs states[] =
    {
        D3DRS_AMBIENT,                  m_ambient,
        D3DRS_COLORVERTEX,              m_color_vertex,
        D3DRS_LIGHTING,                 m_lighting,
        D3DRS_LOCALVIEWER,              m_local_viewer,
        D3DRS_NORMALIZENORMALS,         m_normalize_normals,
        D3DRS_AMBIENTMATERIALSOURCE,    m_ambient_source,
        D3DRS_DIFFUSEMATERIALSOURCE,    m_diffuse_source,
        D3DRS_SPECULARMATERIALSOURCE,   m_specular_source,
        D3DRS_EMISSIVEMATERIALSOURCE,   m_emissive_source,
        D3DRS_SPECULARENABLE,           m_specular_enable,
        D3DRS_SHADEMODE,                m_shade_mode,
        D3DRS_FILLMODE,                 m_fill_mode,
        D3DRS_ALPHATESTENABLE,          false,
        D3DRS_STENCILENABLE,            false,
        D3DRS_ZENABLE,                  D3DZB_TRUE,
        D3DRS_ZFUNC,                    D3DCMP_LESS,
        D3DRS_ZWRITEENABLE,             true,
        D3DRS_ALPHABLENDENABLE,         false,
        D3DRS_DITHERENABLE,             m_dithering
    };
    rt::set_states(m_pd3dDevice, states, NUM_OF(states));

    // orient and scale the teapot
    D3DXMATRIX world = rt::mat_rot_x(m_fWorldRotX)*
        rt::mat_rot_y(m_fWorldRotY)*
        rt::mat_scale(1.4f);
    THR(m_pd3dDevice->SetTransform(D3DTS_WORLD, &world));
    THR(m_teapot->DrawSubset(0));

    // show statistics if necessary
    if (m_show_text)
    {
        RenderText();
    }

    THR(m_pd3dDevice->EndScene());

    return S_OK;
}
